
import { copy, ensureDirSync, emptyDirSync} from 'fs-extra';
import {resolve} from 'path';
import { existsSync } from 'fs';

export const template = /* html*/`
<div id="app">
    <div class="content">
        <h1><ui-label value="i18n:xr-update-engine.title" /></h1>
        <div>
            <ui-button class="update"><ui-label value="i18n:xr-update-engine.update" /></ui-button>
            <ui-button class="restore"><ui-label value="i18n:xr-update-engine.resotre" /></ui-button>
        </div>
        <div class="tip"></div>
    </div>
</div>`;
export const style = `
#app {
    width: 100%;
    height: 100%;
    display: flex;
    align-items: center;
    justify-content: center;
}
.content {
    text-align: center;
}
h1 {
    margin-bottom: 30px;
}
ui-button + ui-button {
    margin-left: 20px;
}
.tip {
    
    font-size: 11px;
    margin-top: 20px;
}
`;
let panel;
let isOparation = false;

export const $ = {
    root: '#app',
    update: '.update',
    restore: '.restore',
    tip: '.tip',
};

export const methods = {
    async update() {
        if (isOparation) return;
        panel.showTip();

        const engine = await Editor.Message.request('engine', 'query-info');
        const pluginPath = await Editor.Package.getPath('xr-update-engine');
        const backupEnginePath = resolve(pluginPath, './engine-backup');

        const isCustom = !(await Editor.Profile.getConfig('engine', `engine.${Editor.App.version.replace(/\./g, '')}.javascript.builtin`));
        if (isCustom || engine.path === backupEnginePath) {
            panel.showTip(panel.t('isCustom'));
            return;
        }
        panel.toggleButtonDisable(true);

        panel.showTip(panel.t('operate'));
        console.time('备份引擎');
        
        ensureDirSync(backupEnginePath);
        emptyDirSync(backupEnginePath); 
        
        try {
            await copy(engine.path, resolve(pluginPath, './engine-backup'));
        } catch (e) {
            panel.showTip(panel.t('operateError'));
            console.error(e);
            panel.toggleButtonDisable(false);
            return;
        }
        console.timeEnd('备份引擎');

        // 解压
        const xrEngineZip = resolve(pluginPath, 'engine-xr.zip');
        if (!existsSync(xrEngineZip)) {
            console.error(`${xrEngineZip} 文件不存在！`);
            panel.toggleButtonDisable(false);
            return;
        }
        await Editor.Utils.File.unzip(xrEngineZip, pluginPath);
        
        // 合并
        const xrEnginePath = resolve(pluginPath, './engine-xr');
        await copy(xrEnginePath, backupEnginePath);

        // 直接将配置存入local
        await Editor.Profile.setConfig('engine', `engine.${Editor.App.version.replace(/\./g, '')}`, {
            javascript: {
                builtin: false,
                custom: backupEnginePath,
            },
            native: {
                builtin: false,
                custom: resolve(backupEnginePath, './native'),
            },
        }, 'local');

        panel.showTip(panel.t('successUpdate'));
        panel.toggleButtonDisable(false);
    },
    async restore() {
        if (isOparation) return;
        panel.showTip();
        panel.toggleButtonDisable(true);

        // 删除存入local的配置
        await Editor.Profile.removeConfig('engine', `engine.${Editor.App.version.replace(/\./g, '')}`, 'local' );

        panel.toggleButtonDisable(false);
        panel.showTip(panel.t('successRestore'));
    },
    showTip(text = '') {
        panel.$.tip.innerText = text;
    },
    toggleButtonDisable(disabled) {
        isOparation = disabled;
        const fn = disabled ? 'setAttribute' : 'removeAttribute';
        panel.$.update[fn]('disabled', disabled);
        panel.$.restore[fn]('disabled', disabled);
    },
    t(key) {
        const name = `xr-update-engine.${key}`;
        return Editor.I18n.t(name);
    },
};

export function ready() {
    panel = this;
    this.$.update.addEventListener('click', this.update, false);
    this.$.restore.addEventListener('click', this.restore, false);
}

export function close() {
    this.$.update.removeEventListener('click', this.update, false);
    this.$.restore.removeEventListener('click', this.restore, false);
}

export function beforeClose() {
    if (isOparation) {
        Editor.Dialog.info(panel.t('operate'));
        return false;
    }
    return true;
}
