/*
 Copyright (c) 2022 Xiamen Yaji Software Co., Ltd.

 https://www.cocos.com

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated engine source code (the "Software"), a limited,
  worldwide, royalty-free, non-assignable, revocable and non-exclusive license
 to use Cocos Creator solely to develop games on your target platforms. You shall
  not use Cocos Creator software for developing other software or tools that's
  used for developing games. You are not granted to publish, distribute,
  sublicense, and/or sell copies of Cocos Creator.

 The software or tools in this License Agreement are licensed, not sold.
 Xiamen Yaji Software Co., Ltd. reserves all rights not expressly granted to you.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 */

/**
 * @packageDocumentation
 * @module component/xr
 */
const { property } = _decorator;
import { _decorator, Component, ccenum, Node, XrEventHandle, Collider, XrControlEventType } from 'cc';
import { IXrInteractable } from './xr-interactable';
import { InteractorEvents } from '../event/interactor-events';

export enum SelectActionTrigger_Type {
    State = 0,
    State_Change = 1,
    Toggle = 2,
    Sticky = 3
}

ccenum(SelectActionTrigger_Type)

export class XrInteractor extends Component {
    @property({ serializable: true })
    protected _attachTransform: Node | null = null;
    @property({ serializable: true })
    protected _selectActionTrigger: SelectActionTrigger_Type = SelectActionTrigger_Type.State;

    protected _triggerState: boolean = false;
    protected _activateTriggerState: boolean = false;
    protected _stateState: boolean = false;
    protected _activateStateState: boolean = false;
    protected _interactorEvents: InteractorEvents | null = null;
    protected _event = new XrEventHandle("XrInteractor");
    protected _collider: Collider | null = null;
    protected _activateCollider: Collider | null = null;
    protected _uiPressCollider: Collider | null = null;
    protected _accupyLine: boolean = false;

    // The triggered object Interactable
    protected _beTriggerNode: IXrInteractable | null = null;

    @property({
        type: Node,
        displayOrder: 1,
        tooltip: 'i18n:xr-plugin.xr_interactor.attachTransform'
    })
    set attachTransform(val) {
        if (val === this._attachTransform) {
            return;
        }
        this._attachTransform = val;
    }
    get attachTransform() {
        return this._attachTransform;
    }

    @property({
        type: SelectActionTrigger_Type,
        displayOrder: 12,
        tooltip: 'i18n:xr-plugin.xr_interactor.selectActionTrigger'
    })
    set selectActionTrigger(val) {
        if (val === this._selectActionTrigger) {
            return;
        }
        this._selectActionTrigger = val;
    }
    get selectActionTrigger() {
        return this._selectActionTrigger;
    }

    set event(val) {
        if (val === this._event) {
            return;
        }
        this._event = val;
    }
    get event() {
        return this._event;
    }

    protected _judgeHit(type: XrControlEventType) {
        return false;
    }

    protected _judgeTrigger() {
        // Determine if the object is captured
        if (!this._beTriggerNode) {
            return false;
        }
        // Has captured the object, judge the captured object, its grasp is its own
        if (this._beTriggerNode.triggerId === this.uuid) {
            return true;
        }
        return false;
    }

    protected _emitSelectEntered(type: XrControlEventType) {
        if (this._event) {
            this._interactorEvents?.selectEntered(this._event);
        }
        if (type === XrControlEventType.SELECT_ENTERED) {
            this._collider?.emit(type, this._event);
        } else if (type === XrControlEventType.ACTIVATED) {
            this._activateCollider?.emit(type, this._event);
        }
        
    }

    private _emitSelectEnd(type: XrControlEventType) {
        if (this._event) {
            this._interactorEvents?.selectExited(this._event);
        }
        if (type === XrControlEventType.SELECT_EXITED) {
            this._collider?.emit(type, this._event);
            this._collider = null;
        } else if (type === XrControlEventType.DEACTIVITED) {
            this._activateCollider?.emit(type, this._event);
            this._activateCollider = null;
        }
    }

    public selectStart(event: XrEventHandle) {
        this._event.model = event.model;
        this._event.eventHandle = event.eventHandle;
        switch (this._selectActionTrigger) {
            case SelectActionTrigger_Type.State:
                this._stateState = true;
                break;
            case SelectActionTrigger_Type.State_Change:
                if (!this._judgeTrigger()) {
                    if (this._judgeHit(XrControlEventType.SELECT_ENTERED)) {
                        this._triggerState = true;
                        this._emitSelectEntered(XrControlEventType.SELECT_ENTERED);
                    }
                }
                break;
            case SelectActionTrigger_Type.Toggle:
                if (this._triggerState && this._judgeTrigger()) {
                    this._emitSelectEnd(XrControlEventType.SELECT_EXITED);
                } else {
                    if (this._judgeHit(XrControlEventType.SELECT_ENTERED)) {
                        this._triggerState = true;
                        this._emitSelectEntered(XrControlEventType.SELECT_ENTERED);
                    }
                }
                break;
            case SelectActionTrigger_Type.Sticky:
                if (this._judgeTrigger()) {
                    this._triggerState = true;
                } else {
                    if (this._judgeHit(XrControlEventType.SELECT_ENTERED)) {
                        this._emitSelectEntered(XrControlEventType.SELECT_ENTERED);
                    }
                }
                break;
            default:
                break;
        }
    }

    public selectEnd(event: XrEventHandle) {
        this._event.model = event.model;
        this._event.eventHandle = event.eventHandle;
        switch (this._selectActionTrigger) {
            case SelectActionTrigger_Type.State:
                if (this._triggerState) {
                    this._emitSelectEnd(XrControlEventType.SELECT_EXITED);
                    this._triggerState = false;
                }
                this._stateState = false;
                break;
            case SelectActionTrigger_Type.State_Change:
                if (this._triggerState) {
                    this._emitSelectEnd(XrControlEventType.SELECT_EXITED);
                    this._triggerState = false;
                }
                break;
            case SelectActionTrigger_Type.Toggle:
                break;
            case SelectActionTrigger_Type.Sticky:
                if (this._triggerState) {
                    this._emitSelectEnd(XrControlEventType.SELECT_EXITED);
                    this._triggerState = false;
                }
                break;
            default:
                break;
        }
    }

    public activateStart(event: XrEventHandle) {
        this._event.model = event.model;
        this._event.eventHandle = event.eventHandle;
        switch (this._selectActionTrigger) {
            case SelectActionTrigger_Type.State:
                this._activateStateState = true;
                break;
            case SelectActionTrigger_Type.State_Change:
                if (!this._judgeTrigger()) {
                    if (this._judgeHit(XrControlEventType.ACTIVATED)) {
                        this._activateTriggerState = true;
                        this._emitSelectEntered(XrControlEventType.ACTIVATED);
                    }
                }
                break;
            case SelectActionTrigger_Type.Toggle:
                if (this._activateTriggerState && this._judgeTrigger()) {
                    this._emitSelectEnd(XrControlEventType.DEACTIVITED);
                } else {
                    if (this._judgeHit(XrControlEventType.ACTIVATED)) {
                        this._activateTriggerState = true;
                        this._emitSelectEntered(XrControlEventType.ACTIVATED);
                    }
                }
                break;
            case SelectActionTrigger_Type.Sticky:
                if (this._judgeTrigger()) {
                    this._activateTriggerState = true;
                } else {
                    if (this._judgeHit(XrControlEventType.ACTIVATED)) {
                        this._emitSelectEntered(XrControlEventType.ACTIVATED);
                    }
                }
                break;
            default:
                break;
        }
    }

    public activateEnd(event: XrEventHandle) {
        this._event.model = event.model;
        this._event.eventHandle = event.eventHandle;

        switch (this._selectActionTrigger) {
            case SelectActionTrigger_Type.State:
                if (this._activateTriggerState) {
                    this._emitSelectEnd(XrControlEventType.DEACTIVITED);
                    this._activateTriggerState = false;
                }
                this._activateStateState = false;
                break;
            case SelectActionTrigger_Type.State_Change:
                if (this._activateTriggerState) {
                    this._emitSelectEnd(XrControlEventType.DEACTIVITED);
                    this._activateTriggerState = false;
                }
                break;
            case SelectActionTrigger_Type.Toggle:
                break;
            case SelectActionTrigger_Type.Sticky:
                if (this._activateTriggerState) {
                    this._emitSelectEnd(XrControlEventType.DEACTIVITED);
                    this._activateTriggerState = false;
                }
                break;
            default:
                break;
        }
    }

    public uiPressEnter(event: XrEventHandle) {
        this._event.deviceType = event.deviceType;
        this._uiPressCollider?.emit(XrControlEventType.UIPRESS_ENTERED, this._event);
    }

    public uiPressExit(event: XrEventHandle) {
        this._event.deviceType = event.deviceType;
        this._uiPressCollider?.emit(XrControlEventType.UIPRESS_EXITED, this._event);
    }
}