/*
 Copyright (c) 2022 Xiamen Yaji Software Co., Ltd.

 https://www.cocos.com

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated engine source code (the "Software"), a limited,
  worldwide, royalty-free, non-assignable, revocable and non-exclusive license
 to use Cocos Creator solely to develop games on your target platforms. You shall
  not use Cocos Creator software for developing other software or tools that's
  used for developing games. You are not granted to publish, distribute,
  sublicense, and/or sell copies of Cocos Creator.

 The software or tools in this License Agreement are licensed, not sold.
 Xiamen Yaji Software Co., Ltd. reserves all rights not expressly granted to you.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 */

import { _decorator, XrEventHandle, XrControlEventType, Component, Collider, XrUIPressEvent, BoxCollider, UITransform, ERigidBodyType, XrUIPressEventType, sys, CCBoolean } from 'cc';
const { ccclass, help, menu, property, executeInEditMode } = _decorator;


/**
 * @en Check XR UI collide with raycast
 * @zh XRUI射线碰撞检测
 */
@ccclass('cc.RaycastChecker')
@help('i18n:cc.RaycastChecker')
@menu('XR/XRUI/RaycastChecker')
@executeInEditMode
export class RaycastChecker extends Component {
    @property({ serializable: true })
    private _ignoreReversedUI = false;

    @property({
        type: CCBoolean,
        displayOrder: 1,
        tooltip: 'i18n:xr-plugin.raycast_checker.ignoreReversedUI'
    })
    set ignoreReversedUI(val) {
        if (val === this._ignoreReversedUI) {
            return;
        }
        this._ignoreReversedUI = val;
    }
    get ignoreReversedUI() {
        return this._ignoreReversedUI;
    }

    private _collider: Collider | null = null;
    private _event: XrUIPressEvent = new XrUIPressEvent("XrUIPressEvent");

    onLoad() {
        if (this.node.getComponent(Collider)) {
            return;
        }
        const collider = this.node.addComponent(BoxCollider);
        const uiTransform = this.node.getComponent(UITransform);
        if (uiTransform) {
            collider.center.set(0, 0, 0);
            collider.size.set(uiTransform.width, uiTransform.height, 0.01);
        }  
    }

    onEnable() {
        this._collider = this.node.getComponent(Collider);
        if (this._collider) {
            this._collider.on(XrControlEventType.HOVER_ENTERED, this._hoverEnter, this);
            this._collider.on(XrControlEventType.HOVER_EXITED, this._hoverExit, this);
            this._collider.on(XrControlEventType.HOVER_STAY, this._hoverStay, this);
            this._collider.on(XrControlEventType.UIPRESS_ENTERED, this._uiPressEnter, this);
            this._collider.on(XrControlEventType.UIPRESS_EXITED, this._uiPressExit, this);
        }
        if (this._collider?.attachedRigidBody) {
            this._collider.attachedRigidBody.type = ERigidBodyType.KINEMATIC;
            this._collider.attachedRigidBody.useGravity = false;
        }
    }

    onDisable() {
        if (this._collider) {
            this._collider.off(XrControlEventType.HOVER_ENTERED, this._hoverEnter, this);
            this._collider.off(XrControlEventType.HOVER_EXITED, this._hoverExit, this);
            this._collider.off(XrControlEventType.HOVER_STAY, this._hoverStay, this);
            this._collider.off(XrControlEventType.UIPRESS_ENTERED, this._uiPressEnter, this);
            this._collider.off(XrControlEventType.UIPRESS_EXITED, this._uiPressExit, this);
        }
    }

    update() {
        // Update node position synchronize collision box position (Note: only update parent node position does not synchronize collision box position)
        if (sys.isXR) {
            this.node.setWorldPosition(this.node.worldPosition);
            this.node.setWorldRotation(this.node.worldRotation);
        }
    }
    
    private _hoverEnter(event: XrEventHandle) {
        this._event.deviceType = event.deviceType;
        this._event.hitPoint.set(event.hitPoint);
        this.node.emit(XrUIPressEventType.XRUI_HOVER_ENTERED, this._event);
    }

    private _hoverExit() {
        this.node.emit(XrUIPressEventType.XRUI_HOVER_EXITED, this);
    }

    private _hoverStay(event: XrEventHandle) {
        this._event.deviceType = event.deviceType;
        this._event.hitPoint.set(event.hitPoint);
        this.node.emit(XrUIPressEventType.XRUI_HOVER_STAY, this._event);
    }

    private _uiPressEnter(event: XrEventHandle) {
        this._event.deviceType = event.deviceType;
        this._event.hitPoint.set(event.hitPoint);
        this.node.emit(XrUIPressEventType.XRUI_CLICK, this._event);
    }

    private _uiPressExit(event: XrEventHandle) {
        this.node.emit(XrUIPressEventType.XRUI_UNCLICK, event.hitPoint);
    }
}